#!/usr/bin/python
# -*- coding: utf-8 -*-

import io
from os import *
from tkinter.filedialog import *

from befehle import *
from compiler import *
from errorWindow import *
from outputWindow import *
from ramregister import *


class MainWindow(tk.Frame):
    __ramreg = RamRegister()

    def __init__(self, parent=None):
        tk.Frame.__init__(self, parent)

        self.frame = None
        self.master = parent
        self.currentFile = None
        self.commandList = None
        self.registerList = None
        self.debugWindow = None
        self.editor = None
        self.outWindow = None

        self.setGUI()
        self.setContent()

    def setGUI(self):
        """
        Erzeugt das Hauptfenster
        """
        # =================== Fenstereigenschaften ===================
        self.frame = tk.Frame(self)
        self.frame.pack(fill="x", side="top", padx="5", pady="5")

        # =================== Menu ===================================
        menu = Menu(self.frame)

        # DateiMenu
        file = Menu(menu)
        file.add_command(
            label="Neu",
            underline=0,
            accelerator="Ctrl+N",
            command=self.newFile)

        file.add_command(
            label="Öffnen",
            underline=0,
            accelerator="Ctrl+O",
            command=self.loadFile)
                    
        file.add_separator()

        file.add_command(
            label="Speichern",
            underline=0,
            accelerator="Ctrl+S",
            command=self.save)

        file.add_command(
            label="Speichern unter",
            underline=0,
            accelerator="F12",
            command=self.saveFileAs)

        file.add_separator()
        file.add_command(label="Beenden", command=exit, underline=0)

        # StartMenu
        start = Menu(menu)
        start.add_command(
            label="Starten",
            underline=0,
            accelerator="F5",
            command=self.startProgram)

        # Hotkeys
        self.bind_all("<Control-n>", self.newFile)
        self.bind_all("<Control-o>", self.loadFile)
        self.bind_all("<Control-s>", self.save)
        self.bind_all("<F5>", self.startProgram)
        self.bind_all("<F12>", self.saveFileAs)

        # MenuBar
        menu.add_cascade(label="Datei", menu=file, underline=0)
        menu.add_cascade(label="Starten", menu=start, underline=0)
        self.master.config(menu=menu)

    def setContent(self):
        """
        Fuegt den Editor und Befehlsbaum in einem getrennten Fenster zusammen
        """

        # Tabreiter fuer verschiedenen Fenster (Editor / Ramregister)
        tabBar = ttk.Notebook(self.master)

        # Editor
        splitView = PanedWindow(tabBar, orient="horizontal")
        splitView.pack(expand=True, fill="both")

        splitView.add(self.buildCommandList())
        splitView.add(self.buildEditor())

        # Ramregister
        registerTab = self.buildRamRegisterList()

        # Verknuepfung mit der Tableiste
        tabBar.add(splitView, text='Quellcode')
        tabBar.add(registerTab, text='Registervorbelegung')
        tabBar.pack(expand=True, fill="both")

    def buildEditor(self):
        """ 
        Erstellt den Editor
        """
        # Erstellung des Editors
        frameEditor = tk.Frame()
        frameEditor.pack(expand=True, fill="both")
                
        sbx = tk.Scrollbar(frameEditor, orient="horizontal")
        sbx.pack(fill="x", side="bottom")

        sby = tk.Scrollbar(frameEditor)
        sby.pack(fill="y", side="right")

        self.editor = tk.Text(frameEditor)
        self.editor.config(wrap="none")  # wrap="word"  word char 
        self.editor.pack(expand=True, fill="both", padx="5", pady="5")

        self.editor["xscrollcommand"] = sbx.set
        sbx["command"] = self.editor.xview
        self.editor["yscrollcommand"] = sby.set
        sby["command"] = self.editor.yview

        return frameEditor

    def buildCommandList(self):
        """ 
        Erstellt den Befehlsbaum im linken Bereich des Editors
        """
        commandList = tk.Frame()
        commandList.pack(expand=True, fill="both", side="top", padx="5", pady="5")

        ysb = tk.Scrollbar(commandList)
        ysb.pack(fill="y", side="right")

        xsb = tk.Scrollbar(commandList, orient = "horizontal")
        xsb.pack(fill="x", side="bottom")
                
        tree = tk.ttk.Treeview(commandList)
        tree.pack(fill="both", expand=True)
        tree["xscrollcommand"] = xsb.set
        xsb["command"] = tree.xview
        ysb["command"] = tree.yview

        # Eintragen der Befehle in den Befehlsbaum für den Editor
        tree.heading('#0', text="Befehle", anchor='w')

        befehleDictionary = Befehle.getBefehleAsDictionary(Befehle)
        id = 0
        for key, val in sorted(befehleDictionary.items()):
            try:
                id = tree.insert("", "end", str(key), text=str(key))
            except:
                pass

            for befehl in val:
                tree.insert(id, "end", text=befehl)

            tree.bind("<Double-1>", self.doCommand)
            self.commandList = tree

        return commandList

    def doCommand(self, event):
        """ Wird ausgefuehrt, wenn ein Befehl in den Editor eingetragen werden soll """
        if self.commandList is None:
            return

        command = self.commandList.identify("item", event.x, event.y)

        if command:
            self.editor.insert("end", self.commandList.item(command)["text"])
            self.editor.insert("end", "\n")


    def buildRamRegisterList(self):
        """ 
        Baut den Tab fuer die Ramregistervorbelgung
        """		
        registerList = tk.Frame()
        registerList.pack(expand=True, fill="both", side="top", padx="5", pady="5")

        ysb = tk.Scrollbar(registerList)
        ysb.pack(fill="y", side="right")

        xsb = tk.Scrollbar(registerList, orient = "horizontal")
        xsb.pack(fill="x", side="bottom")
                
        tree = tk.ttk.Treeview(registerList, columns=('col0', 'col1', 'col2'))
        tree.pack(fill="both", expand=True)
        tree["xscrollcommand"] = xsb.set
        xsb["command"] = tree.xview
        ysb["command"] = tree.yview

        # Eintragen der Register
        tree['show'] = 'headings'
        tree.column("col0", width=20, anchor="center")
        tree.column("col1", width=20, anchor="center")
        tree.column("col2", width=100, anchor="center")
        tree.heading("col0", text='Adr')
        tree.heading("col1", text='Hex-Wert')
        tree.heading("col2", text='Name')

        # Eintragen der Registerwerte
        for i in range(Konstanten.MAX_REGISTER):
            tree.insert("", "end", values=(str(i), str(0), ""))

        tree.bind("<Double-1>", self.edit)

        self.registerList = tree
        return registerList

    def edit(self, event):
        """ Wird ausgefuehrt, wenn eine Ramregisterzelle manuell bearbeitet wird """
        result = tk.simpledialog.askinteger("Neuer Registerwert", "Bitte geben Sie einen neuen "
                                                                  "Zahlenwert für dieses Register ein.")
        if result:
            item = self.registerList.selection()[0]

        if item:
            self.registerList.item(item, values=(str(self.registerList.index(item)), str(result),
                                                   self.registerList.item(item, "values")[2]))
                        
        result = tk.simpledialog.askstring("Neuer Name", "Bitte geben Sie einen neuen Namen für dieses Register ein.")
        if result:
            item = self.registerList.selection()[0]

        if item:
            self.registerList.item(item, values=(str(self.registerList.index(item)),
                                                   self.registerList.item(item, "values")[1], result))

    def newFile(self, event=None):
        """ 
        Editor für neue Datei leeren
        """
        # Editor bereinigen
        self.editor.delete("1.0", "end")
        self.currentFile = None

    def loadFile(self, event=None):
        """ 
        Laedt eine vorhandene CPUX-Datei in den Editor
        """
        file = askopenfile(mode='r', defaultextension=".cpux",
                           filetypes=(("Resimdateien (*.cpux)", "*.cpux"), ("All files", "*.*")))

        if file:
            try:
                # Datei einlesen
                fileContent = file.readlines()
                versionNumber = fileContent[0].split(" ")

                if int(versionNumber[0]) != Konstanten.VERSION:
                    messagebox.showinfo("Falsche Version", "Die Datei wurde mit einer anderen Versionsnummer erstellt!")
                    return

                # Register einlesen
                items = self.registerList.get_children()
                for i in range(0, Konstanten.MAX_REGISTER):
                    item = items[i]
                    self.registerList.item(item, values=(str(i+1), fileContent[i+1].split("\t")[1],
                                                         self.registerList.item(item, "values")[2]))

                # Editor bereinigen
                self.editor.delete("1.0", "end")
                self.currentFile = file.name

                for i in range(Konstanten.MAX_REGISTER + 1, len(fileContent)):
                    self.editor.insert("end", fileContent[i])
                            
                #file.close()
            except:                     
                messagebox.showwarning("Öffnen der Datei", "Kann die Datei nicht öffnen! \n(%s)" % self.currentFile)
            finally:
                file.close()

    def save(self, event=None):
        """ 
        Speichert die Datei unter dem bereits vorhandenen Namen oder legt einen neuen fest, wenn keiner vorhanden ist
        """	
        if self.currentFile is None:
            self.saveFileAs()
        else:
            file = None
            try:
                file = io.open(self.currentFile, 'w')

                # Programm in Datei abspeichern
                file.write("%d Version\n" % Konstanten.VERSION)
                #os.write(fd, str("%d Version\n" % Konstanten.VERSION))

                # Register speichern
                cntRegister = 1
                for reg in self.registerList.get_children():
                    #print(self.registerList.item(child)["values"])
                    file.write("%d\t%d\n" % (cntRegister, (self.registerList.item(reg)["values"])[1]))

                    #os.write(fd, str("%d\t%d\n" % (cntRegister, (self.registerList.item(reg)["values"])[1])))
                    cntRegister += 1

                file.write(self.editor.get('1.0', END+'-1c'))
                #os.write(fd, str(self.editor.get('1.0', END+'-1c')))
            except:
                messagebox.showwarning("Speichern der Datei", "Beim Speichern der Datei ist ein Fehler aufgetreten!")
            finally:
                if file is not None:
                    file.close()

                #if fd > -1:
                #    os.close(fd)
                    
    def saveFileAs(self, event=None):
        """ 
        Speichert die Datei unter dem ausgewaehlten Pfad
        """
        file = asksaveasfile(mode='w', defaultextension=".cpux", filetypes=(("Resimdateien (*.cpux)",
                                                                             "*.cpux"), ("All Files", "*.*")))

        if file:
            try:
                self.currentFile = file.name

                # Programm in Datei abspeichern
                file.write("%d Version\n" % Konstanten.VERSION)

                # Register speichern
                cntRegister = 1
                for child in self.registerList.get_children():
                    #print(self.registerList.item(child)["values"])
                    file.write("%d\t%d\n" % (cntRegister, (self.registerList.item(child)["values"])[1]))
                    cntRegister += 1

                file.write(self.editor.get('1.0', END+'-1c'))
                #file.close()
            except:
                messagebox.showwarning("Speichern der Datei", "Beim Speichern der Datei ist ein Fehler aufgetreten!")
            finally:
                file.close()

    def startProgram(self, event=None):
        """ 
        Startet das selbstgeschriebene Resim-Programm in einem neuen Fenster
        """
        self.__ramreg.clear()
        self.save()

        compiler = Compiler()
        ram = compiler.checkSourceCode(self.editor.get('1.0', END+'-1c'))
        error = False
        errortext = ""

        for bef in ram:
            if bef is None:
                continue

            if bef.error:
                error = True
                errortext = errortext + "\r\n\r\n" + bef.oriline + "\r\n  " + str(bef.errorText)

        # Test auf doppelte Marker
        if not error:
            for i in range(0, len(ram)):
                bef1 = ram[i]

                for j in range(0, len(ram)):
                    bef2 = ram[j]

                    if i != j and len(bef1.markertext) > 0 and \
                                    len(bef2.markertext) > 0 and bef1.markertext == bef2.markertext:
                        error = True
                        errortext = errortext + "\r\n\r\n" + bef2.oriline + "\r\n  Doppelte Sprungmarken"

        # Ersetzen der Zielsprungmarken auf Adressen
        if not error:
            for i in range(0, len(ram)):
                bef1 = ram[i]
                if len(bef1.zielmarkertext) > 0:
                    markervorhanden = False

                    for j in range(0, len(ram)):
                        bef2 = ram[j]
                        if i != j and bef1.zielmarkertext == bef2.markertext:
                            markervorhanden = True
                            bef1.jumpnr = j
                            bef1.konstante = j

                    if not markervorhanden:
                        error = True
                        if len(errortext) == 0:
                            errortext = bef1.oriline + "\r\n  Die Spruingmarke ist nicht vorhanden"
                        else:
                            errortext = errortext + "\r\n\r\n" + bef1.oriline + "\r\n  " \
                                                                                "Die Sprungmarke ist nicht vorhanden"

        if error:
            errorWindow = ErrorWindow(self, errortext, os.path.basename(self.currentFile) if
                                      self.currentFile is not None else "Nicht gespeicherte Datei")
            return
        else:
            for bef in ram:
                self.__ramreg.addCommand(bef)

        if self.__ramreg.countRam() == 0:
            return

        # Register
        regName = ""
        value = 0
        ordernr = 0

        regs = self.registerList.get_children()
        for row in range(0, len(regs)):
            value = (self.registerList.item(regs[row])["values"])[1]
            ordernr = value
            regName = (self.registerList.item(regs[row])["values"])[2]

            if ordernr > 0:
                self.__ramreg.setRegister(row, None, ordernr, regName)
            else:
                self.__ramreg.setRegister(row, None, None, regName)

        if self.outWindow is None:
            self.outWindow = OutputWindow(self, self.__ramreg)

    def on_closing(self, event=None):
        self.outWindow = None
