#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import tkinter as tk
from tkinter.simpledialog import *
from tkinter.ttk import *
from tkinter import messagebox

from ramregister import *
from befehl import *
from steuerwerk import *
from resultNextStep import *
from ausgabeWindow import *
from EAEinheit import *
from optionsWindow import *


class OutputWindow(tk.Frame):
    __ramreg = None
    __debug = False
    __lastDebugIP = -1
    __liescOk = True
    __stepMode = False
    __programSpeed = 300
    __isRunning = False

    __outea = 0
    __outakku = 0
    __outoperand = 0
    __outIP = 0
    __zFlag = 0
    __cFlag = 0
    __pFlag = 0
    __sFlag = 0

    def __init__(self, parent, ramreg):
        tk.Frame.__init__(self, parent)
        self.frame = tk.Toplevel(parent)
        self.master = parent
        self.frame.protocol("WM_DELETE_WINDOW", self.close)

        self.__ramreg = ramreg

        self.memory = None
        self.register = None
        self.toolbar = None
        self.ausgabe = None

        self.__outea = StringVar()
        self.__outakku = StringVar()
        self.__outoperand = StringVar()
        self.__zFlag = StringVar()
        self.__cFlag = StringVar()
        self.__pFlag = StringVar()
        self.__sFlag = StringVar()
        self.__outIP = StringVar()

        self.__outea.set(str(hex(0)))
        self.__outakku.set(str(hex(0)))
        self.__outoperand.set(str(hex(0)))
        self.__zFlag.set(str(0))
        self.__cFlag.set(str(0))
        self.__pFlag.set(str(0))
        self.__sFlag.set(str(0))
        self.__outIP.set(str(hex(0)))

        self.setGUI()
        self.setProgram()

    def setGUI(self):
        """
         Initialisiert die GUI des Debugfensters
        """
        # =================== Fenstereigenschaften ===================
        self.frame.wm_title("Simulationsprogramm Resim")
        # Immer oberstes Fenster
        self.frame.wm_attributes("-topmost", 1)

        self.frame.grid_rowconfigure(0, weight=0)
        self.frame.grid_rowconfigure(1, weight=10)
        self.frame.grid_rowconfigure(2, weight=10)
        self.frame.grid_rowconfigure(3, weight=10)

        self.frame.grid_columnconfigure(0, weight=1)
        self.frame.grid_columnconfigure(1, weight=10)
        self.frame.grid_columnconfigure(2, weight=1)

        # Toolbar
        self.toolbar = Frame(self.frame, borderwidth=2, relief="raised")
        self.toolbar.grid_columnconfigure(0, weight=0)
        self.toolbar.grid_columnconfigure(1, weight=0)
        self.toolbar.grid_columnconfigure(2, weight=0)
        self.toolbar.grid_columnconfigure(3, weight=0)

        # =================== Menu ===================================
        menu = Menu(self.frame)

        # DateiMenu
        file = Menu(menu)
        file.add_command(label="Beenden", command=self.frame.destroy, underline=0)

        ansicht = Menu(menu)
        ansicht.add_command(
            label="Ausgabefenster",
            underline=0,
            command=self.showEA)

        # StartMenu
        start = Menu(menu)
        start.add_command(
            label="Starten von IP=0",
            underline=0,
            accelerator="F5",
            command=lambda: self.startProgram(False, 0, False))

        start.add_separator()
        start.add_command(
            label="Debug von IP=0",
            underline=0,
            accelerator="Strg+F5",
            command=lambda: self.startProgram(True, 0, False))

        start.add_separator()
        start.add_command(
            label="Starten - Einzelschritt",
            underline=0,
            command=lambda: self.startProgram(False, 0, True))

        start.add_command(
            label="Einzelschritt",
            underline=0,
            accelerator="F10",
            command=self.nextStep)

        start.add_command(
            label="Stop",
            underline=0,
            command=self.stopProgram)

        extras = Menu(menu)
        extras.add_command(
            label="Taschenrechner",
            underline=0,
            command=lambda: os.system('start calc.exe'))

        extras.add_separator()
        extras.add_command(
            label="Optionen",
            underline=0,
            command=self.showOptions)

        # Hotkeys
        self.bind_all("<F5>", lambda: self.startProgram(False, 0, False))
        self.bind_all("<Control-F5>", lambda: self.startProgram(True, 0, False))
        self.bind_all("<F10>", self.nextStep)

        # MenuBar
        menu.add_cascade(label="Datei", menu=file, underline=0)
        menu.add_cascade(label="Ansicht", menu=ansicht, underline=0)
        menu.add_cascade(label="Starten", menu=start, underline=0)
        menu.add_cascade(label="Extras", menu=extras, underline=0)

        self.frame.config(menu=menu)

        btn_start = tk.Button(self.toolbar, text="Start von Adr. 0", fg="red", font="Verdana 11 bold",
                              command=lambda: self.startProgram(False, 0, False))
        btn_ip = tk.Button(self.toolbar, text="Start bis Haltepunkt (Debug)", fg="red", font="Verdana 11 bold",
                           command=lambda: self.startProgram(True, 0, False))
        btn_start_step = tk.Button(self.toolbar, text="Start Step", fg="red", font="Verdana 11 bold",
                                   command=lambda: self.startProgram(False, 0, True))
        btn_step = tk.Button(self.toolbar, text="Step", fg="red", font="Verdana 11 bold",
                             command=self.nextStep)
        btn_stop = tk.Button(self.toolbar, text="Stop", fg="red", font="Verdana 11 bold",
                             command=self.stopProgram)

        btn_start.grid(row=0, column=0)
        btn_ip.grid(row=0, column=1)
        btn_start_step.grid(row=0, column=2)
        btn_step.grid(row=0, column=3)
        btn_stop.grid(row=0, column=4)

        self.toolbar.grid(row=0, column=0, columnspan=3)

        # Rechenwerk
        rechenwerkLabel = LabelFrame(self.frame, text="Rechenwerk")

        akkuLabel = Label(rechenwerkLabel, text="Akku")
        akkuValue = Label(rechenwerkLabel, textvariable=self.__outakku)

        optLabel = Label(rechenwerkLabel, text="Operand")
        optValue = Label(rechenwerkLabel, textvariable=self.__outoperand)

        zFlagLabel = Label(rechenwerkLabel, text="Z-Flag")
        zFlagValue = Label(rechenwerkLabel, textvariable=self.__zFlag)

        cFlagLabel = Label(rechenwerkLabel, text="C-Flag")
        cFlagValue = Label(rechenwerkLabel, textvariable=self.__cFlag)

        pFlagLabel = Label(rechenwerkLabel, text="P-Flag")
        pFlagValue = Label(rechenwerkLabel, textvariable=self.__pFlag)

        sFlagLabel = Label(rechenwerkLabel, text="S-Flag")
        sFlagValue = Label(rechenwerkLabel, textvariable=self.__sFlag)

        akkuLabel.grid(row=0, column=0)
        akkuValue.grid(row=1, column=0)

        optLabel.grid(row=0, column=1)
        optValue.grid(row=1, column=1)

        zFlagLabel.grid(row=2, column=0)
        zFlagValue.grid(row=3, column=0)
        cFlagLabel.grid(row=2, column=1)
        cFlagValue.grid(row=3, column=1)
        pFlagLabel.grid(row=2, column=2)
        pFlagValue.grid(row=3, column=2)
        sFlagLabel.grid(row=2, column=3)
        sFlagValue.grid(row=3, column=3)

        rechenwerkLabel.grid(row=1, column=2, padx=10, pady=10)

        # EA-Einheit
        eaLabel = LabelFrame(self.frame, text="E/A Einheit")

        inputBox = Label(eaLabel, textvariable=self.__outea)
        inputBox.grid(row=0, column=0, padx=10, pady=10)

        eaLabel.grid(row=1, column=0, sticky='w', padx=10, pady=10)

        # Speicher
        memoryLabel = LabelFrame(self.frame, text="Speicher")

        self.memory = self.buildMemory()
        if self.memory is not None:
            self.memory.grid(row=2, column=0, rowspan=2, columnspan=2, sticky='nsew', padx=10, pady=10)

        memoryLabel.grid(row=2, column=0, padx=10, pady=10)

        # Steuerwerk
        strLabel = LabelFrame(self.frame, text="Steuerwerk")

        ipFlag = Label(strLabel, text="IP")
        ipValue = Label(strLabel, textvariable=self.__outIP)

        ipFlag.grid(row=0, column=0)
        ipValue.grid(row=0, column=1)

        strLabel.grid(row=2, column=2, padx=10, pady=10)

        # Register
        registerLabel = LabelFrame(self.frame, text="Register")

        self.register = self.buildRamRegisterList()
        if self.register is not None:
            self.register.grid(row=3, column=2, sticky='nsew', padx=10, pady=10)

        registerLabel.grid(row=3, column=2, padx=10, pady=10)

    def buildMemory(self):
        """
        Zeigt die Befehle im Speicher an
        """
        ysb = tk.Scrollbar(self.frame)
        xsb = tk.Scrollbar(self.frame, orient="horizontal")

        tree = tk.ttk.Treeview(self.frame, columns=('col0', 'col1', 'col2', 'col3', 'col4'), selectmode="browse")
        tree["xscrollcommand"] = xsb.set
        xsb["command"] = tree.xview
        ysb["command"] = tree.yview

        # Eintragen der Register
        tree['show'] = 'headings'
        tree.column("col0", width=50, anchor="center", stretch=NO)
        tree.column("col1", width=75, anchor="center", stretch=NO)
        tree.column("col2", width=100, anchor="center", stretch=NO)
        tree.column("col3", width=150, anchor="center")
        tree.column("col4", width=150, anchor="center")

        tree.heading("col0", text='Debug')
        tree.heading("col1", text='Adr')
        tree.heading("col2", text='Hex-Wert')
        tree.heading("col3", text='Befehl')
        tree.heading("col4", text='Kommentar')

        tree.bind("<Double-1>", self.editMemory)

        return tree

    def buildRamRegisterList(self):
        """
        Baut den Tab fuer die Ramregister
        """
        ysb = tk.Scrollbar(self.frame)
        xsb = tk.Scrollbar(self.frame, orient="horizontal")

        tree = tk.ttk.Treeview(self.frame, columns=('col0', 'col1', 'col2'))
        tree["xscrollcommand"] = xsb.set
        xsb["command"] = tree.xview
        ysb["command"] = tree.yview

        # Eintragen der Register
        tree['show'] = 'headings'
        tree.column("col0", width=20, anchor="center")
        tree.column("col1", width=20, anchor="center")
        tree.column("col2", width=100, anchor="center")
        tree.heading("col0", text='Adr')
        tree.heading("col1", text='Hex-Wert')
        tree.heading("col2", text='Name')

        # Eintragen der Registerwerte
        #for i in range(20):
        #    tree.insert("", "end", values=(str(i), str(0), ""))

        #tree.bind("<Double-1>", self.editRegister)

        return tree

    def editMemory(self, event=None):
        """ Wird ausgefuehrt, wenn eine Kom manuell bearbeitet wird """
        item = self.memory.selection()[0]

        debugmarke = 'x'
        if item:
            if self.memory.item(item, "values")[0] == 'x':
                debugmarke = ''

            self.memory.item(item, values=(str(debugmarke), self.memory.item(item, "values")[1],
                                                self.memory.item(item, "values")[2],
                                                self.memory.item(item, "values")[3],
                                                self.memory.item(item, "values")[4]))

            self.__ramreg.getBefehl(int(self.memory.item(item, "values")[1])).debug = (debugmarke == 'x')

    # aktuell nicht benutzt
    def editRegister(self, event=None):
        """ Wird ausgefuehrt, wenn eine Kom manuell bearbeitet wird """
        result = tk.simpledialog.askinteger("Neuer Befehl", "Bitte geben Sie einen neuen "
                                                            "Befehl für diesen Speicher ein.")
        if result:
            item = self.memory.selection()[0]

        if item:
            self.memory.item(item, values=(str(self.memory.index(item)), str(result),
                                           self.memory.item(item, "values")[2]))

        result = tk.simpledialog.askstring("Neuer Kommentar", "Bitte geben Sie einen neuen Kommentar "
                                                              "für diesen Speicher ein.")
        if result:
            item = self.memory.selection()[0]

        if item:
            self.memory.item(item, values=(str(self.memory.index(item)),
                                           self.memory.item(item, "values")[3], result))

    def setProgram(self):
        bef = None
        order = ""
        row = 0

        for i in range(0, self.__ramreg.countRam()):
            bef = self.__ramreg.getBefehl(i)
            if bef is None:
                break

            order = bef.oriline
            if bef.ordernr == Konstanten.C_IN_A_LIESC:
                if len(bef.liesZahlText) == 0:
                    order = "LiesZahl "
                else:
                    order = "LiesZahl " + "\"" + bef.liesZahlText + "\""
            elif bef.ordernr == Konstanten.C_OUT_STRING:
                order += bef.ausgabeString
            elif bef.ordernr in [Konstanten.C_Move_Akku_Const,
                                 Konstanten.C_Add_Akku_Const,
                                 Konstanten.C_Sub_Akku_Const,
                                 Konstanten.C_Mod_Akku_Const,
                                 Konstanten.C_Mult_Akku_Const,
                                 Konstanten.C_Div_Akku_Const,
                                 Konstanten.C_And_Akku_Const,
                                 Konstanten.C_Or_Akku_Const,
                                 Konstanten.C_Xor_Akku_Const,
                                 Konstanten.C_SHL_Akku_Const,
                                 Konstanten.C_SHR_Akku_Const,
                                 Konstanten.C_ROL_Akku_Const,
                                 Konstanten.C_ROR_Akku_Const,
                                 Konstanten.C_CMP_Akku_Const]:
                order += " (" + str(hex(bef.konstante)) + "h)"
            elif bef.ordernr in [Konstanten.C_JZ,
                                 Konstanten.C_JNZ,
                                 Konstanten.C_JS,
                                 Konstanten.C_JNS,
                                 Konstanten.C_JL,
                                 Konstanten.C_JLE,
                                 Konstanten.C_JG,
                                 Konstanten.C_JGE,
                                 Konstanten.C_JC,
                                 Konstanten.C_JNC,
                                 Konstanten.C_JP,
                                 Konstanten.C_JNP,
                                 Konstanten.C_JUMP]:
                order += " " + str(bef.konstante)
            #elif bef.ordernr in [Konstanten.C_Move_Akku_Reg,
            #                     Konstanten.C_Add_Akku_Reg,
            #                     Konstanten.C_Sub_Akku_Reg,
            #                     Konstanten.C_Mod_Akku_Reg,
            #                     Konstanten.C_Mult_Akku_Reg,
            #                     Konstanten.C_Div_Akku_Reg,
            #                     Konstanten.C_And_Akku_Reg,
            #                     Konstanten.C_Or_Akku_Reg,
            #                     Konstanten.C_Xor_Akku_Reg,
            #                     Konstanten.C_SHL_Akku_Reg,
            #                     Konstanten.C_SHR_Akku_Reg,
            #                     Konstanten.C_CMP_Akku_Reg,
            #                     Konstanten.C_OUT_REG,
            #                     Konstanten.C_OUT_REGHex,
            #                     Konstanten.C_IN_REG]:
            #    order += " " + str(bef.register)

            # Ram schreiben
            self.memory.insert("", "end", iid="Row %s" % i, values=("", i, hex(bef.ordernr), order, bef.comment))

        # Register schreiben
        for i in range(0, Konstanten.MAX_REGISTER):
            reg = self.__ramreg.getRegister(i)

            if reg is not None:
                self.register.insert("", "end", iid="Reg %s" % i, values=(str(i), hex(reg.number), reg.comment))

    def stopProgram(self):
        self.__isRunning = False

    def startProgram(self, isDebug, startIP, stepMode):
        if self.__isRunning:
            return

        self.__debug = isDebug
        self.__stepMode = stepMode
        self.__isRunning = True

        self.checkEAAusgabe()
        if self.ausgabe is not None:
            self.ausgabe.clear()

        sw = Steuerwerk()
        sw.setParent(self)
        sw.setIP(startIP)

        self.memory.selection_set("Row " + str(startIP))
        self.after(self.__programSpeed, self.nextStep)

    def checkEAAusgabe(self):
        if self.ausgabe is None:
            for i in range(0, self.__ramreg.countRam()):
                bef = self.__ramreg.getBefehl(i)
                if bef is not None:
                    if self.indexof("OUT", bef.oriline.upper()) > -1 or \
                                    self.indexof("IN", bef.oriline.upper()) > -1 or \
                                    self.indexof("LIES", bef.oriline.upper()) > -1:
                        self.showEA()
                        break

    def nextStep(self, event=None):
        if not self.__isRunning:
            return

        sw = Steuerwerk()
        liesc = sw.checkLiesc()

        if liesc:
            liescOk = True

        ip = sw.getIP()

        if self.__debug and self.__lastDebugIP != ip:
            self.__lastDebugIP = ip
            if ip < self.__ramreg.countRam():
                bef = self.__ramreg.getBefehl(ip)
                if bef.debug:
                    return

        self.__lastDebugIP = ip
        res = sw.nextStep()

        # Endzustaende:
        #   1. Division durch 0
        #   2. Programm ist am Ende
        if res == ResultNextStep.DIV0:
            messagebox.showerror("Division durch 0", "Sie haben versucht durch 0 zu dividieren!")
            self.__stepMode = False
            self.__isRunning = False
            return

        if res == ResultNextStep.ENDE:
            self.__stepMode = False
            self.__isRunning = False
            return

        # naechste zeile setzen
        self.updateGUI_RAM(ip)
        self.memory.selection_set("Row " + str(ip))

        if not self.__stepMode:
            self.after(self.__programSpeed, self.nextStep)

    def showOptions(self, event=None):
        opDialog = OptionsDialog(self.frame)
        self.__programSpeed = opDialog.result

    def on_closing(self):
        self.ausgabe = None

    def showEA(self, event=None):
        if self.ausgabe is None:
            self.ausgabe = AusgabeWindow(self)

    def updateGUI_EA(self, ausgabe=None):
        eaEinheit = EAEinheit()
        value = None

        if ausgabe is None:
            value = eaEinheit.getEAValue()

            if 0 <= value <= Konstanten.MAX_WERT:
                self.__outea.set(str(hex(value)))
                self.frame.update_idletasks()
        else:
            value = ausgabe

        if self.ausgabe is not None:
            self.ausgabe.add(str(value))
            self.ausgabe.add("\r\n")

    def updateGUI_NL(self):
        if self.ausgabe is not None:
            self.ausgabe.add("\r\n")

    def updateGUI_Akku_Operand(self):
        rw = Rechenwerk()

        #self.__outakku.set(str(hex(rw.getAkku())))
        #self.__outoperand.set(str(hex(rw.getOperand())))

        self.__outakku.set(str(hex((rw.getAkku() + (1 << 32)) % (1 << 32))))
        self.__outoperand.set(str(hex((rw.getOperand() + (1 << 32)) % (1 << 32))))

        self.__zFlag.set(str(self.boolToInt(rw.isZeroFlag())))
        self.__cFlag.set(str(self.boolToInt(rw.isCeroFlag())))
        self.__sFlag.set(str(self.boolToInt(rw.isSignFlag())))
        self.__pFlag.set(str(self.boolToInt(rw.isParityFlag())))

        self.update_idletasks()

    def boolToInt(self, integer):
        return 1 if integer else 0

    def updateGUI_Register(self):
        for row in range(0, Konstanten.MAX_REGISTER):
            reg = self.__ramreg.getRegister(row)
            if reg is not None:
                item = "Reg %s" % row
                self.register.item(item, values=(str(row), hex(reg.number), reg.comment))
                #self.register.insert("", "end", values=(str(i), hex(reg.number), reg.comment))

        self.frame.update_idletasks()

    def updateGUI_Stop(self):
        sw = Steuerwerk()
        messagebox.showerror("Crash", "Fehler beim Ausfuehren des Programms! Fehler: " + str(sw.isCrashed()))
        self.__stepMode = False
        self.__isRunning = False

    def updateGUI_RAM(self, ip):
        self.__outIP.set(str(hex(ip)))
        self.frame.update_idletasks()

    def indexof(self, char, string):
        return string.index(char) if char in string else -1

    def close(self):
        self.master.on_closing()
        self.frame.destroy()
