#!/usr/bin/python
# -*- coding: utf-8 -*-

from konstanten import *


class Rechenwerk(object):
    __instance = None

    __isZeroFlag = False
    __isCarryFlag = False
    __isParityFlag = False
    __isSignFlag = False
    __isOpUndefined = False

    __akku = 0
    __operand = 0

    # Singleton
    def __new__(cls):
        if Rechenwerk.__instance is None:
            Rechenwerk.__instance = object.__new__(cls)

        return Rechenwerk.__instance

    ####################### Getter/Setter #######################
    def getAkku(self):
        return self.__akku

    def setAkku(self, value):
        self.__akku = value
        self.setFlags(value)

    def setOperand(self, value):
        self.__operand = value
        self.__isOpUndefined = False

    def getOperand(self):
        return 0 if self.__isOpUndefined else self.__operand

    def isZeroFlag(self):
        return self.__isZeroFlag

    def setZeroFlag(self, value):
        self.__isZeroFlag = value

    def isCeroFlag(self):
        return self.__isCarryFlag

    def setCarryFlag(self, value):
        self.__isCarryFlag = value

    def isParityFlag(self):
        return self.__isParityFlag

    def setParityFlag(self, value):
        cnt = 0
        while value > 0:
            if value % 2 > 0:
                cnt += 1
            value = value >> 1

        if cnt % 2 == 0:
            self.__isParityFlag = True
        else:
            self.__isParityFlag = False

    def isSignFlag(self):
        return self.__isSignFlag

    def setSignFlag(self, value):
        self.__isSignFlag = value

    def clearCarry(self):
        self.__isCarryFlag = False

    def isOpUndefined(self):
        return self.__isOpUndefined

    def setOpUndefined(self, value):
        self.__isOpUndefined = True
        self.__operand = 0

    def writeAkkuToOperand(self):
        self.__operand = self.__akku

    def setFlags(self, value):
        self.__isZeroFlag = (value == 0)
        self.__isSignFlag = False
        self.setParityFlag(value)


    ####################### Arithmetische Operationen #######################
    def addition(self):
        summe = self.__akku + self.__operand

        if summe > Konstanten.MAX_WERT:
            self.setAkku((summe - (Konstanten.MAX_WERT + 1)))
            self.__isCarryFlag = True
        else:
            self.setAkku(summe)
            self.__isCarryFlag = False

        return self.__akku

    def subtraktion(self):
        if self.__akku > self.__operand:
            self.setAkku((Konstanten.MAX_WERT + 1) + self.__operand - self.__akku)
            self.__isCarryFlag = True
        else:
            self.setAkku(self.__operand - self.__akku)
            self.__isCarryFlag = False

        return self.__akku

    def multiplikation(self):
        zwerg = self.__akku * self.__operand

        if zwerg > Konstanten.MAX_WERT:
            self.setAkku(zwerg - (Konstanten.MAX_WERT + 1))
            self.__isCarryFlag = True
        else:
            self.setAkku(zwerg)
            self.__isCarryFlag = False

        return self.__akku

    def division(self):
        divzero = False

        if self.__akku == 0:
            divzero = True
        else:
            if self.__akku > self.__operand:
                self.setAkku(0)
                self.__isCarryFlag = False
            else:
                self.setAkku(self.__operand // self.__akku)
                self.__isCarryFlag = False

        return divzero

    ####################### Logische Operationen #######################
    def procAnd(self):
        self.setAkku(self.__akku & self.__operand)
        return self.__akku

    def moduloRegister(self):
        self.setAkku(self.__operand % self.__akku)
        return self.__akku

    def modulo(self):
        self.setAkku(self.__operand % self.__akku)
        self.__isCarryFlag = False
        return self.__akku

    def divRegister(self):
        divzero = False

        if self.__akku == 0:
            divzero = True
        else:
            self.setAkku(self.__operand // self.__akku)

        return divzero, self.__akku

    def procDec(self):
        self.setAkku(self.__akku - 1)
        return self.__akku

    def procInc(self):
        if self.__akku >= Konstanten.MAX_WERT:
            self.setAkku(((self.__akku + 1) % (Konstanten.MAX_WERT + 1)))
            self.__isCarryFlag = True
        else:
            self.setAkku(self.__akku + 1)
            self.__isCarryFlag = False

        return self.__akku

    def procDaa(self):
        value = self.__akku
        carry = 0
        shift = 1
        zwerg = 0

        while value > 0:
            value += carry
            rest = value % 16
            if rest < 10:
                zwerg += rest * shift
                carry = 0
            else:
                rest += 6
                rest -= 16
                zwerg += rest * shift
                carry = 1

            shift *= 10
            value = (value // 16)

        self.setAkku(zwerg)
        self.__isCarryFlag = (carry == 1)

        return self.__akku

    def procDas(self):
        value = self.__akku
        carry = 0
        shift = 1
        zwerg = 0

        while value > 0:
            value += carry
            rest = value % 16
            if rest < 10:
                zwerg += rest * shift
                carry = 0
            else:
                rest -= 6
                zwerg += rest * shift
                carry = 1

            shift *= 10
            value = (value // 16)

        self.setAkku(zwerg)
        self.__isCarryFlag = (carry == 1)

        return self.__akku

    def procOr(self):
        self.setAkku((self.__akku | self.__operand))
        self.__isCarryFlag = False

        return self.__akku

    def procXor(self):
        self.setAkku((self.__akku ^ self.__operand))
        self.__isCarryFlag = False

        return self.__akku

    def procNot(self):
        self.setAkku(~self.__akku)
        self.__isCarryFlag = False

        return self.__akku

    def procRol(self):
        msb = 0
        i = 0
        op = self.__akku
        akku = self.__operand

        for zhl in range(1, op + 1):
            msb = 1 if self.__isCarryFlag else 0
            i = akku & Konstanten.MAX_WERT_ROL
            self.__isCarryFlag = (i == Konstanten.MAX_WERT_ROL)

            akku = (akku << 1)
            akku = (akku | msb)

        self.setAkku(akku)
        return self.__akku

    def procRor(self):
        lsb = 0
        i = 0
        op = self.__akku
        akku = self.__operand

        for zhl in range(1, op + 1):
            i = akku & 1
            lsb = 1 if self.__isCarryFlag else 0
            self.__isCarryFlag = (i == 1)

            akku = (akku >> 1)
            akku = (akku | (lsb << 7))

        self.setAkku(akku)
        return self.__akku

    def procShl(self):
        i = 0
        zhl = 0
        op = self.__akku
        akku = self.__operand

        for zhl in range(1, op + 1):
            i = akku % 128
            self.__isCarryFlag = (i == 128)
            akku = (akku << 1)

        self.setAkku(akku)
        return self.__akku

    def procShr(self):
        i = 0
        zhl = 0
        op = self.__akku
        akku = self.__operand

        for zhl in range(1, op + 1):
            i = akku % 128
            self.__isCarryFlag = (i == 128)
            akku = (akku >> 1)

        self.setAkku(akku)
        return self.__akku

    def procCMP(self):
        zwerg = self.__operand - self.__akku
        if zwerg >= 0:
            self.setAkku(zwerg)
            self.__isCarryFlag = False
            self.__isSignFlag = False
        else:
            ergFF = zwerg
            ergFF = (ergFF & 0xFFFF)
            self.setAkku(ergFF)
            self.__isCarryFlag = False
            self.__isSignFlag = True

        return self.__akku

    def procJump(self, befehl):
        bZ = self.__isZeroFlag
        bS = self.__isSignFlag
        bC = self.__isCarryFlag
        bP = self.__isParityFlag
        bTest = not bZ
        self.__isCarryFlag = False

        if befehl == Konstanten.C_JZ:
            return bZ
        elif befehl == Konstanten.C_JNZ:
            return not bZ
        elif befehl == Konstanten.C_JS:
            return bS
        elif befehl == Konstanten.C_JNS:
            return not bS
        elif befehl == Konstanten.C_JL:
            return bS
        elif befehl == Konstanten.C_JLE:
            return bS | bZ
        elif befehl == Konstanten.C_JG:
            return (not bS) & (not bZ)
        elif befehl == Konstanten.C_JGE:
            return (not bS) | bZ
        elif befehl == Konstanten.C_JC:
            return bC
        elif befehl == Konstanten.C_JNC:
            return not bC
        elif befehl == Konstanten.C_JP:
            return bP
        elif befehl == Konstanten.C_JNP:
            return not bP

        return False
