#!/usr/bin/python
# -*- coding: utf-8 -*-

from ramregister import *
from resultNextStep import *
from rechenwerk import *
from inputWindow import *
from EAEinheit import *

from tkinter.filedialog import *


class Steuerwerk:
    __instance = None
    __parent = None

    __ramreg = RamRegister()
    __rechenwerk = Rechenwerk()
    __eaEinheit = EAEinheit()
    __ip = 0
    __crash = 0

    # Singleton
    def __new__(cls):
        if Steuerwerk.__instance is None:
            Steuerwerk.__instance = object.__new__(cls)

        return Steuerwerk.__instance

    def setParent(self, parent):
        self.__parent = parent

    def initIP(self):
        self.__ip = Konstanten.IP_START_ADRESSE

    def setIP(self, value):
        if value != Konstanten.UNDEFINIERT:
            self.__ip = value

    def getIP(self):
        return self.__ip

    def getBefehl(self):
        if 0 <= self.__ip <= Konstanten.MAX_RAM:
            return self.__ramreg.getBefehl(self.__ip)
        else:
            return None

    def setCrashed(self, value):
        self.__crash = value

    def isCrashed(self):
        return self.__crash

    def checkLiesc(self):
        bef = self.getBefehl()
        if bef is None:
            return False

        return bef.ordernr == Konstanten.C_IN_A_LIESC

    def nextStep(self):
        bef = self.getBefehl()
        if bef is None:
            return ResultNextStep.ENDE

        if bef.ordernr == Konstanten.CPU_STOP:
            self.setCrashed(Konstanten.ERR_CPU_STOP)

        elif bef.ordernr == Konstanten.C_Move_Akku_Const:
            self.__rechenwerk.setAkku(bef.konstante)
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_Move_Akku_Reg:
            self.__rechenwerk.setAkku(self.__ramreg.getRegister(bef.register).number)
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_Move_Reg_Akku:
            self.__ramreg.getRegister(bef.register).number = self.__rechenwerk.getAkku()
            self.__parent.updateGUI_Register()

        elif bef.ordernr == Konstanten.C_Add_Akku_Reg:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(self.__ramreg.getRegister(bef.register).number)
            self.__rechenwerk.addition()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_Add_Akku_Const:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(bef.konstante)
            self.__rechenwerk.addition()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_Sub_Akku_Reg:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(self.__ramreg.getRegister(bef.register).number)
            self.__rechenwerk.subtraktion()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_Sub_Akku_Const:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(bef.konstante)
            self.__rechenwerk.subtraktion()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_Mod_Akku_Reg:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(self.__ramreg.getRegister(bef.register).number)
            self.__rechenwerk.moduloRegister()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_Mod_Akku_Const:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(bef.konstante)
            self.__rechenwerk.modulo()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_Mult_Akku_Const:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(bef.konstante)
            self.__rechenwerk.multiplikation()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_Div_Akku_Reg:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(self.__ramreg.getRegister(bef.register).number)

            divzero = False
            divzero = self.__rechenwerk.divRegister()

            if divzero:
                return ResultNextStep.DIV0
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_Div_Akku_Const:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(bef.konstante)

            divzero = False
            divzero = self.__rechenwerk.division()

            if divzero:
                return ResultNextStep.DIV0
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_INC:
            self.__rechenwerk.procInc()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_DEC:
            self.__rechenwerk.procDec()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_DAA:
            self.__rechenwerk.procDaa()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_DAS:
            self.__rechenwerk.procDas()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_And_Akku_Reg:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(self.__ramreg.getRegister(bef.register).number)
            self.__rechenwerk.procAnd()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_And_Akku_Const:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(bef.konstante)
            self.__rechenwerk.procAnd()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_Or_Akku_Reg:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(self.__ramreg.getRegister(bef.register).number)
            self.__rechenwerk.procOr()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_Or_Akku_Const:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(bef.konstante)
            self.__rechenwerk.procOr()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_Xor_Akku_Reg:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(self.__ramreg.getRegister(bef.register).number)
            self.__rechenwerk.procXor()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_Xor_Akku_Const:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(bef.konstante)
            self.__rechenwerk.procXor()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_Not_Akku:
            self.__rechenwerk.procNot()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_SHL_Akku_Reg:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(self.__ramreg.getRegister(bef.register).number)
            self.__rechenwerk.procShl()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_SHL_Akku_Const:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(bef.konstante)
            self.__rechenwerk.procShl()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_SHR_Akku_Reg:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(self.__ramreg.getRegister(bef.register).number)
            self.__rechenwerk.procShr()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_SHR_Akku_Const:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(bef.konstante)
            self.__rechenwerk.procShr()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_ROL_Akku_Const:
            self.__rechenwerk.operand = self.__rechenwerk.getAkku()
            self.__rechenwerk.setAkku(bef.konstante)
            self.__rechenwerk.procRol()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_ROR_Akku_Const:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(bef.konstante)
            self.__rechenwerk.procRor()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_CMP_Akku_Reg:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(self.__ramreg.getRegister(bef.register).number)
            self.__rechenwerk.procCMP()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_CMP_Akku_Const:
            self.__rechenwerk.setOperand(self.__rechenwerk.getAkku())
            self.__rechenwerk.setAkku(bef.konstante)
            self.__rechenwerk.procCMP()
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_JUMP:
            self.__ip = bef.konstante
            self.__parent.updateGUI_RAM(self.__ip)
            return ResultNextStep.OKAY

        elif bef.ordernr in [Konstanten.C_JZ, Konstanten.C_JNZ, Konstanten.C_JS, Konstanten.C_JNS,
                             Konstanten.C_JL, Konstanten.C_JLE, Konstanten.C_JG, Konstanten.C_JGE,
                             Konstanten.C_JC, Konstanten.C_JNC, Konstanten.C_JP, Konstanten.C_JNP]:

            bJump = self.__rechenwerk.procJump(bef.ordernr)

            if bJump:
                self.__ip = bef.konstante
                self.__parent.updateGUI_RAM(self.__ip)
                return ResultNextStep.OKAY

        elif bef.ordernr == Konstanten.C_IN_A_LIESC:
            #input = tk.simpledialog.askinteger("Eingabe", "Bitte geben Sie eine Zahl zum Einlesen ein!",
            #                                   parent=self.__parent)
            input = InputWindow(self.__parent).getResult()
            if input is not None and input > -1:
                self.__eaEinheit.setEAValue(input)
                self.__parent.updateGUI_EA()
            else:
                self.__parent.updateGUI_Stop()
                self.setCrashed(Konstanten.ERR_CPU_STOP)

        elif bef.ordernr == Konstanten.C_OUT_STRING:
            self.__parent.updateGUI_EA(bef.ausgabeString)

        elif bef.ordernr == Konstanten.C_IN_Akku:
            self.__rechenwerk.setAkku(self.__eaEinheit.getEAValue())
            self.__parent.updateGUI_Akku_Operand()

        elif bef.ordernr == Konstanten.C_OUT_Akku:
            self.__eaEinheit.setEAValue(self.__rechenwerk.getAkku())
            self.__parent.updateGUI_EA()

        elif bef.ordernr == Konstanten.C_OUT_AkkuHex:
            self.__eaEinheit.setEAValue(self.__rechenwerk.getAkku())
            self.__parent.updateGUI_EA()

        elif bef.ordernr == Konstanten.C_IN_REG:
            self.__ramreg.getRegister(bef.register).number = self.__eaEinheit.getEAValue()
            self.__parent.updateGUI_Register()

        elif bef.ordernr in [Konstanten.C_OUT_REG, Konstanten.C_OUT_REGHex]:
            self.__eaEinheit.setEAValue(self.__ramreg.getRegister(bef.register).number)
            self.__parent.updateGUI_EA()

        elif bef.ordernr == Konstanten.C_OUT_NL:
            self.__parent.updateGUI_NL()

        elif bef.ordernr == Konstanten.C_CLEARREGISTER:
            self.__ramreg.clearRegister()
            self.__parent.updateGUI_Register()

        elif bef.ordernr == Konstanten.C_STOP:
            return ResultNextStep.ENDE

        elif bef.ordernr == Konstanten.C_ClearCarry:
            self.__rechenwerk.clearCarry()
            self.__parent.updateGUI_Akku_Operand()

        self.__ip += 1
        return ResultNextStep.OKAY